import asyncio
import os
import pprint
import sys
from asyncio import subprocess
from pathlib import Path

from rich.text import Text
from rtry import retry

from maxwelld.core.compose_data_types import ServicesComposeState
from maxwelld.core.config import Config
from maxwelld.core.utils.process_command_output import process_output_till_done
from maxwelld.helpers.jobs_result import JobResult
from maxwelld.helpers.jobs_result import OperationError
from maxwelld.output.console import CONSOLE
from maxwelld.output.styles import Style


class ComposeShellInterface:
    def __init__(self, compose_files, in_docker_project_root, execution_envs: dict = None):
        self.compose_files = compose_files
        self.in_docker_project_root = in_docker_project_root
        self.execution_envs = os.environ | {
            'COMPOSE_FILE': self.compose_files,
            'DOCKER_HOST': Config().docker_host,
            'COMPOSE_PROJECT_NAME': Config().compose_project_name,
        }
        if execution_envs is not None:
            self.execution_envs |= execution_envs
        self.verbose_docker_compose_commands = Config().verbose_docker_compose_commands
        self.debug_docker_compose_commands = Config().debug_docker_compose_commands
        self.verbose_docker_compose_ps_commands = Config().verbose_docker_compose_ps_commands
        self.extra_exec_params = Config().docker_compose_extra_exec_params

    @retry(attempts=10, delay=1, until=lambda x: x == JobResult.BAD)
    async def dc_state(self, env: dict = None, root: Path | str = None) -> ServicesComposeState | OperationError:
        sys.stdout.flush()

        if env is None:
            env = {}
        env = self.execution_envs | env

        if root is None:
            root = self.in_docker_project_root

        process = await asyncio.create_subprocess_shell(
            cmd := f"/usr/local/bin/docker-compose --project-directory {root}" + " ps -a --format='{{json .}}'",
            env=env,
            cwd=root,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        CONSOLE.print(Text(
            f'{cmd}',
            style=Style.context
        ))
        stdout, stderr = await process_output_till_done(process, self.verbose_docker_compose_ps_commands)

        if process.returncode != 0:
            print(f"Can't get container's status {stdout} {stderr}")
            return OperationError(f'Stdout:\n{stdout}\n\nStderr:\n{stderr}')

        state_result = ServicesComposeState(stdout.decode('utf-8'))
        CONSOLE.print(Text('Services status result:', style=Style.info))
        CONSOLE.print(state_result.as_rich_text())
        return state_result

    @retry(attempts=3, delay=1, until=lambda x: x == JobResult.BAD)
    async def dc_up(self, services: list[str], env: dict = None, root: Path | str = None) -> JobResult | OperationError:
        sys.stdout.flush()

        if env is None:
            env = {}
        env = self.execution_envs | env

        if root is None:
            root = self.in_docker_project_root

        process = await asyncio.create_subprocess_shell(
            cmd := f'/usr/local/bin/docker-compose --project-directory {root} up --timestamps --no-deps --pull missing '
                   '--timeout 300 -d ' + ' '.join(services),
            env=env,
            cwd=root,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        debug = f'; in {root}; with {pprint.pformat(env)}' if self.debug_docker_compose_commands else ''
        CONSOLE.print(Text(
            f'{cmd}',
            style=Style.context
        ) + ' ' + Text(
            f'{debug}',
            style=Style.regular
        ))
        stdout, stderr = await process_output_till_done(process, self.verbose_docker_compose_commands)

        if process.returncode != 0:
            print("Can't up environment")
            state_result = await self.dc_state()
            if state_result == JobResult.GOOD:
                return OperationError(
                    f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result.as_rich_text()}'
                )
            return OperationError(f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result}')

        return JobResult.GOOD

    @retry(attempts=3, delay=1, until=lambda x: x == JobResult.BAD)
    async def dc_logs(self, services: list[str], env: dict = None, root: Path | str = None, logs_param='--no-log-prefix'
                      ) -> tuple[JobResult, bytes] | tuple[OperationError, None]:
        sys.stdout.flush()

        if env is None:
            env = {}
        env = self.execution_envs | env

        if root is None:
            root = self.in_docker_project_root

        if services is None:
            services = []
        services = ' '.join(services)

        process = await asyncio.create_subprocess_shell(
            cmd := f'/usr/local/bin/docker-compose --project-directory {root} logs {logs_param} {services}',
            env=env,
            cwd=root,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        CONSOLE.print(Text(
            f'{cmd}',
            style=Style.context
        ))
        stdout, stderr = await process_output_till_done(process, False)

        if process.returncode != 0:
            print(f"Can't get {services} logs")
            state_result = await self.dc_state()
            if state_result == JobResult.GOOD:
                return OperationError(
                    f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result.as_rich_text()}'
                ), None
            return OperationError(f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result}'), None

        return JobResult.GOOD, stdout

    @retry(attempts=3, delay=1, until=lambda x: x == JobResult.BAD)
    async def dc_exec(self, container: str, cmd: str, env: dict = None, root: Path | str = None
                      ) -> tuple[JobResult, bytes, bytes] | tuple[OperationError, bytes, bytes]:
        print(f'Executing {cmd} in {container} container')
        sys.stdout.flush()

        if env is None:
            env = {}
        env = self.execution_envs | env

        if root is None:
            root = self.in_docker_project_root

        process = await asyncio.create_subprocess_shell(
            cmd := f'/usr/local/bin/docker-compose --project-directory {root} exec {self.extra_exec_params} {container} {cmd}',
            env=env,
            cwd=root,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        debug = f'; in {root}; with {env}' if self.debug_docker_compose_commands else ''
        CONSOLE.print(Text(
            f'{cmd}',
            style=Style.context
        ) + ' ' + Text(
            f'{debug}',
            style=Style.regular
        ))
        stdout, stderr = await process_output_till_done(process, self.verbose_docker_compose_commands)

        if process.returncode != 0:
            print(f"Can't execute {cmd} in {container} successfully:\n{stdout=}, {stderr=}")
            state_result = await self.dc_state()
            if state_result == JobResult.GOOD:
                return OperationError(
                    f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result.as_rich_text()}'
                ), stdout, stderr
            return OperationError(
                f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result}'
            ), stdout, stderr

        return JobResult.GOOD, stdout, stderr

    @retry(attempts=3, delay=1, until=lambda x: x == JobResult.BAD)
    async def dc_down(self, services: list[str], env: dict = None,
                      root: Path | str = None) -> JobResult | OperationError:
        print(f'Downing {services} containers')
        sys.stdout.flush()

        if env is None:
            env = {}
        env = self.execution_envs | env

        if root is None:
            root = self.in_docker_project_root

        process = await asyncio.create_subprocess_shell(
            cmd := f'/usr/local/bin/docker-compose --project-directory {root} down ' + ' '.join(services),
            env=env,
            cwd=root,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        debug = f'; in {root}; with {env}' if self.debug_docker_compose_commands else ''
        CONSOLE.print(Text(
            f'{cmd}',
            style=Style.context
        ) + ' ' + Text(
            f'{debug}',
            style=Style.regular
        ))
        stdout, stderr = await process_output_till_done(process, self.verbose_docker_compose_commands)

        if process.returncode != 0:
            # TODO swap print to CONSOLE
            print(f"Can't down {services} successfully")
            state_result = await self.dc_state()
            if state_result == JobResult.GOOD:
                return OperationError(
                    f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result.as_rich_text()}'
                )
            return OperationError(f'Stdout:\n{stdout}\n\nStderr:\n{stderr}\n\nComposeState:\n{state_result}')

        return JobResult.GOOD



import asyncio
from typing import TypedDict

from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.client.types import EnvironmentId
from maxwelld.core.service import MaxwellDemonServiceManager
from maxwelld.helpers.bytes_pickle import base64_pickled

UP_LOCK = asyncio.Lock()


class StatusRequestParams(TypedDict):
    id: EnvironmentId


class StatusResponseParams(TypedDict):
    status: str


async def http_get_status(request: Request) -> web.Response:
    params: StatusRequestParams = await request.json()
    status = await MaxwellDemonServiceManager().get().status(env_id=params['id'])
    return web.json_response(StatusResponseParams(status=base64_pickled(status)), status=200)



"""
This module used for tests-side Environment description.

Environments could be predefined with services sets and services special variables.

Environment class used as up command param, describes enabled services and it's envs.
"""



import os
import shutil
from copy import deepcopy
from pathlib import Path
from typing import Union

from rich import json

from maxwelld.client.types import EnvironmentId
from maxwelld.core.sequence_run_types import EMPTY_ID
from maxwelld.core.sequence_run_types import EnvInstanceConfig
from maxwelld.core.utils.compose_instance_cfg import make_env_instance_config
from maxwelld.env_description.env_types import Environment
from maxwelld.helpers.bytes_pickle import base64_pickled
from maxwelld.helpers.bytes_pickle import debase64_pickled


class InflightKeeper:
    def __init__(self, tmp_envs_path: Path, env_file_name: str):
        self.tmp_envs_path = tmp_envs_path
        self.env_file_name = env_file_name
        self._started_envs: dict = {}
        self._envs_file_path = self.tmp_envs_path / self.env_file_name

    def cleanup_in_flight(self):
        if os.path.exists(self._envs_file_path):
            with open(self._envs_file_path, 'r') as envs_file:
                envs_ = envs_file.read()
                if not envs_:
                    envs_ = '{}'
                return json.loads(envs_)

        self.tmp_envs_path.mkdir(parents=True, exist_ok=True)
        dirpath, dirnames, filenames = next(os.walk(self.tmp_envs_path))
        for dir in dirnames:
            # TODO execute dc down for each env folder
            # TODO remove file interference (2 jobs on same runner will write/read same file)
            shutil.rmtree(Path(dirpath) / dir)

    def get_existing_inflight_env(self, name: str, config_template: Environment | None,
                                  compose_files: str) -> Union[EnvInstanceConfig, None]:
        if name in self._started_envs:
            env_id = self._started_envs[name]['env_id']
            env_config_instance = make_env_instance_config(
                env_template=config_template,
                env_id=env_id
            )

            # TODO run dc to check if services started, kill remainig if not.
            # compose_files_instance = make_env_compose_instance_files(
            #     env_config_instance,
            #     compose_files,
            #     project_network_name=self._project,
            #     host_project_root_directory=self.host_project_root_directory,
            #     compose_files_path=self.compose_files_path,
            #     dot_env_files_path=self.dot_env_files_path,
            #     tmp_env_path=self.tmp_envs_path,
            # )
            # make_debug_bash_env(compose_files_instance, self.host_env_tmp_directory)
            #
            return env_config_instance

        return None

    def update_inflight_envs(self, name, config_template: Environment, compose_files: str, env_id: str):
        self._started_envs[name] = {
            'env_id': env_id,
            'params': {
                'name': name,
                'compose_files': compose_files,
                'config_template': base64_pickled(config_template)
            }
        }

        with open(self._envs_file_path, 'w') as envs_file:
            envs_file.write(
                json.dumps(self._started_envs)
            )  # TODO fix sync when env upped but no record or record exists but bo env

    def check_env_limits(self) -> list[dict]:
        in_flight = deepcopy(self._started_envs)
        to_delete = []
        for env_name, env_params in in_flight.items():
            if env_params['env_id'] == EMPTY_ID:  # different env starts EMPTY_ID
                # down_in_flight_envs(
                #     self.tmp_envs_path,
                #     EMPTY_ID,
                #     self.in_docker_project_root_path,
                #     except_containers=self._non_stop_containers
                # )
                to_delete += [in_flight[env_name]]
                del self._started_envs[env_name]
        return to_delete

    def get_existing_inflight_env_by_id(self, env_id: EnvironmentId) -> Union[EnvInstanceConfig, None]:
        for env_name in self._started_envs:
            if self._started_envs[env_name]['env_id'] == env_id:
                config_template = debase64_pickled(self._started_envs[env_name]['params']['config_template'])
                env_config_instance = make_env_instance_config(
                    env_template=config_template,
                    env_id=env_id
                )
                return env_config_instance
        return None

    def get_existing_inflight_env_compose_files(self, env_id: EnvironmentId) -> Union[EnvInstanceConfig, None]:
        for env_name in self._started_envs:
            if self._started_envs[env_name]['env_id'] == env_id:
                return self._started_envs[env_name]['params']['compose_files']
        return None



import json
from dataclasses import dataclass
from typing import Callable
from typing import Iterator

from rich.text import Text

from maxwelld.output.styles import Style


class ComposeState:
    RUNNING = 'running'


class ComposeHealth:
    EMPTY = ''
    HEALTHY = 'healthy'


@dataclass
class ServiceComposeState:
    name: str
    state: str
    health: str
    status: str  # "Up X seconds"

    @classmethod
    def from_json(cls, json_status: str) -> 'ServiceComposeState':
        status = json.loads(json_status)
        return cls(
            name=status['Service'],
            state=status['State'],
            health=status['Health'],
            status=status['Status'],
        )

    def __eq__(self, other):
        return (isinstance(other, ServiceComposeState)
                and self.name == other.name
                and self.state == other.state
                and self.health == other.health)

    def __repr__(self):
        return (f'{type(self).__name__}'
                f'(name="{self.name}", '
                f'state="{self.state}", '
                f'health="{self.health}", '
                f'status="{self.status}")')

    def as_rich_text(self, style: Style = Style()):
        service_string = Text('     ')
        service_string.append(Text(f"{self.name:{20}}", style=style.regular))
        service_string.append(Text(
            f"{self.state:{20}}",
            style=style.good if self.state == ComposeState.RUNNING else style.bad
        ))
        service_string.append(Text(
            f"{self.health:{20}}",
            style=style.good if self.health == ComposeHealth.HEALTHY else style.bad
        ))
        service_string.append(Text(
            self.status, style=style.regular
        ))
        service_string.append(Text('\n', style=style.regular))
        return service_string

    def as_json(self) -> dict[str, str]:
        return {
            'name': self.name,
            'state': self.state,
            'health': self.health,
            'status': self.status,
        }


class ServicesComposeState:
    def __init__(self, compose_status: str):
        self._services: list[ServiceComposeState] = [
            ServiceComposeState.from_json(state_str)
            for state_str in compose_status.split('\n')
            if state_str
        ]

    def __contains__(self, item):
        return item in self._services

    def __iter__(self) -> Iterator[ServiceComposeState]:
        return iter(self._services)

    def as_rich_text(
        self,
        filter: Callable[[ServiceComposeState], bool] = lambda x: True,
        style: Style = Style()
    ) -> Text:
        services_text = Text()
        for service_state in self._services:
            if filter(service_state):
                services_text.append(service_state.as_rich_text(style))
        return services_text

    def __eq__(self, other) -> bool:
        if isinstance(other, ServicesComposeState):
            for service_state in self._services:
                if service_state not in other:
                    return False
            for service_state in other:
                if service_state not in self:
                    return False
            return True

        return False

    def __repr__(self):
        return f'{type(self).__name__}(<{self._services}>)'

    def as_json(self, filter: Callable[[ServiceComposeState], bool] = lambda x: True, ) -> list[dict]:
        return [service_status.as_json() for service_status in self._services if filter(service_status)]



from pathlib import Path

from maxwelld.core.sequence_run_types import ComposeInstanceFiles
from maxwelld.core.utils.compose_instance_cfg import get_new_instance_compose_files


def make_debug_bash_env(env_config_compose_instance: ComposeInstanceFiles,
                        host_tmp_env_path: Path):
    new_external_compose_file = get_new_instance_compose_files(
        env_config_compose_instance.compose_files_source,
        host_tmp_env_path / env_config_compose_instance.env_config_instance.env_id
    )

    with open(env_config_compose_instance.directory / '.env', 'w') as f:
        f.write(f'alias dc="docker-compose --project-directory ."\n')
        f.write(f'export COMPOSE_FILE={new_external_compose_file}\n')
        f.write(f'alias deactivate-env="unset COMPOSE_FILE"\n')
        # TODO Envs
        # for k,value in updated_envs.items():
        #     f.write(f'{k}={value}\n')
        # f.write(f'export COMPOSE_PROJECT_NAME={new_project_name}\n')



from typing import TypedDict

from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.client.types import EnvironmentId
from maxwelld.core.service import MaxwellDemonServiceManager
from maxwelld.helpers.bytes_pickle import base64_pickled


class DcExecRequestParams(TypedDict):
    env_id: EnvironmentId
    container: str
    command: str


class DcExecResponseParams(TypedDict):
    output: str


async def dc_exec(request: Request) -> web.Response:
    params: DcExecRequestParams = await request.json()

    output = await MaxwellDemonServiceManager().get().exec(
        env_id=params['env_id'],
        container=params['container'],
        command=params['command']
    )
    return web.json_response(DcExecResponseParams(output=base64_pickled(output)), status=200)



from enum import Enum
from enum import auto
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple

DEFAULT_ENV = 'DEFAULT'


class Environments:
    def __getitem__(self, item) -> 'Environment':
        return getattr(self, item)

    @classmethod
    def list_all(cls) -> list[str]:
        return [
            item
            for item in cls.__dict__
            if not item.startswith('__') or not item.startswith('_')
        ]


class AsIs:
    def __init__(self, value):
        self.value = value


class Env(Dict):
    ...


class StageName(NamedTuple):
    compose_name: str


class EventStage(Enum):
    BEFORE_ALL = StageName('before_all')
    BEFORE_SERVICE_START = StageName('before_start')
    AFTER_SERVICE_START = StageName('after_start')
    AFTER_SERVICE_HEALTHY = StageName('after_healthy')
    AFTER_ALL = StageName('after_all')

    @classmethod
    def get_all_stages(cls):
        return [
            cls.BEFORE_ALL,
            cls.BEFORE_SERVICE_START,
            cls.AFTER_SERVICE_START,
            cls.AFTER_SERVICE_HEALTHY,
            cls.AFTER_ALL,

        ]

    @classmethod
    def get_all_compose_stages(cls):
        return [stage.value.compose_name for stage in cls.get_all_stages()]

    @classmethod
    def get_compose_stage(cls, stage_name: str) -> 'EventStage':
        for stage in cls.get_all_stages():
            if stage.value.compose_name == stage_name:
                return stage
        assert False, 'No such stage: {}'.format(stage_name)


class Handler(NamedTuple):
    stage: EventStage
    cmd: str
    executor: str = None


class ServiceMode(Enum):
    ON = auto()
    OFF = auto()
    SINGLETON = auto()
    EXTERNAL = auto()


class Service(NamedTuple):
    name: str
    env: Env = Env()
    events_handlers: List[Handler] = []
    mode: ServiceMode = ServiceMode.ON

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return f'Service({self.name}, {self.mode})'

    def with_env(self, env: Env):
        return Service(
            name=self.name,
            env=Env(self.env | env),
            events_handlers=self.events_handlers,
            mode=self.mode
        )

    def as_dict(self):
        return {
            'name': self.name,
            'env': dict(self.env),
        }


def remove_dups(*services: Service) -> List[Service]:
    result_services = []
    for service in reversed(services):
        if service not in result_services:
            result_services += [service]

    return list(reversed(result_services))


class Environment:  # TODO rename Environment
    @classmethod
    def from_environment(cls, env: 'Environment', name, *services: Service):
        # TODO duplicated services merging
        return Environment(name, *env._services, *services)

    def __init__(self, name, *services: Service):
        # TODO duplicated services merging
        self._name = name  # TODO think how to extract from envs?
        self._services = remove_dups(*services)
        self._services_dict: dict[str, Service] = {
            item.name: item for item in self._services
        }

    def __str__(self) -> str:
        return self._name

    def __repr__(self):
        return f'Environment({self._name})'

    def get_services(self) -> dict:
        return self._services_dict

    def __getitem__(self, item) -> Service:
        return self._services_dict[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self._services_dict)

    def isidentifier(self):
        return True

    def __eq__(self, other):
        return self._name == other

    def __hash__(self):
        return hash(self._name)

    def as_json(self) -> list[dict]:
        return [
            service.as_dict() for service in self._services
        ]

class SingletonService(Service):
    singleton: bool = False



import asyncio
from typing import TypedDict

from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.client.types import EnvironmentId
from maxwelld.core.service import MaxwellDemonServiceManager
from maxwelld.helpers.bytes_pickle import base64_pickled

UP_LOCK = asyncio.Lock()


class EnvRequestParams(TypedDict):
    id: EnvironmentId


class EnvResponseParams(TypedDict):
    env: str


async def http_get_env(request: Request) -> web.Response:
    params: EnvRequestParams = await request.json()
    env = MaxwellDemonServiceManager().get().env(env_id=params['id'])
    return web.json_response(data=EnvResponseParams(env=base64_pickled(env)), status=200)



from enum import Enum
from enum import auto


class JobResult(Enum):
    GOOD = auto()
    BAD = auto()


class OperationError:
    def __init__(self, log: str):
        self.log = log

    def __eq__(self, other):
        return other == JobResult.BAD

    def __repr__(self):
        return f'Operation finished unsuccessful:\n{self.log}'



import collections
import shutil
from _warnings import warn
from copy import deepcopy
from pathlib import Path

import yaml

from maxwelld.core.sequence_run_types import ComposeInstanceFiles
from maxwelld.core.sequence_run_types import EnvInstanceConfig
from maxwelld.core.utils.compose_instance_cfg import get_new_instance_compose_files
from maxwelld.env_description.env_types import Environment
from maxwelld.env_description.env_types import EventStage
from maxwelld.env_description.env_types import Handler
from maxwelld.errors.migrations import ServicesMigrationsError


def read_dc_file(filename: str | Path) -> dict:
    with open(filename) as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def write_dc_file(filename: str | Path, cfg: dict) -> None:
    with open(filename, 'w') as f:
        f.write(yaml.dump(cfg))


def patch_network(dc_cfg: dict, network_name) -> dict:
    new_dc_cfg = deepcopy(dc_cfg)
    if 'networks' not in new_dc_cfg:
        new_dc_cfg['networks'] = {}
    # new_dc_cfg['networks']['e2e_back_network'] = {
    #     'name': network_name,
    #     'external': True,
    # }

    for service in new_dc_cfg['services']:
        if 'network' not in new_dc_cfg['services'][service]:
            new_dc_cfg['services'][service]['networks'] = []
        # new_dc_cfg['services'][service]['networks'] += ['e2e_back_network']
    return new_dc_cfg


def patch_service_volumes(volumes: list, root_path: str | Path) -> list:
    updated_volumes = []
    for volume in volumes:
        new_volume = volume
        if volume.startswith('.'):
            new_volume = volume.replace('.', str(root_path), 1)
        updated_volumes += [new_volume]
    return updated_volumes


def patch_services_volumes(dc_cfg: dict, root_path: str | Path) -> dict:
    new_dc_cfg = deepcopy(dc_cfg)
    for service in dc_cfg['services']:
        if 'volumes' in dc_cfg['services'][service]:
            updated_volumes = patch_service_volumes(
                dc_cfg['services'][service]['volumes'],
                root_path
            )
            new_dc_cfg['services'][service]['volumes'] = updated_volumes
    return new_dc_cfg


def list_key_exist(key, env: list[str]) -> None | str:
    for item in env:
        if key in item:
            return item
    else:
        return None


def patch_service_set(dc_cfg: dict, services_map: dict[str, str]):
    new_dc_cfg = deepcopy(dc_cfg)
    for service in dc_cfg['services']:
        if service not in services_map:
            del new_dc_cfg['services'][service]
    return new_dc_cfg


def patch_envs(dc_cfg: dict, services_environment_vars: Environment):
    # TODO envs order and override question!!
    #  if we overrides env, should we save order? or insert before, for allow to override codegen
    new_dc_cfg = deepcopy(dc_cfg)
    for service in dc_cfg['services']:
        if 'environment' not in dc_cfg['services'][service]:
            new_dc_cfg['services'][service]['environment'] = []
        if isinstance(new_dc_cfg['services'][service]['environment'], list) and service in services_environment_vars:
            for k, v in services_environment_vars[service].env.items():
                if existing := list_key_exist(f'{k}={v}',
                                              new_dc_cfg['services'][service]['environment']):
                    if existing != f'{k}={v}':
                        warn(
                            f'⚠️ env {k} for service {service} already set to "{existing}" '
                            f'instead of "{v}"')
                else:
                    new_dc_cfg['services'][service]['environment'] += [f'{k}={v}']
        elif isinstance(new_dc_cfg['services'][service]['environment'], dict) and service in services_environment_vars:
            for k, v in services_environment_vars[service].env.items():
                if k in new_dc_cfg['services'][service]['environment']:
                    if new_dc_cfg['services'][service]['environment'][k] != v:
                        warn(f'⚠️ env {k} for service {service} already set to '
                             f"\"{new_dc_cfg['services'][service]['environment'][k]}\" instead "
                             f"of \"{v}\"")
                else:
                    new_dc_cfg['services'][service]['environment'].update({k: v})
    return new_dc_cfg


def patch_services_names(dc_cfg: dict, services_map: dict[str, str]) -> dict:
    new_service_dc_cfg = deepcopy(dc_cfg)
    new_service_dc_cfg['services'] = {}
    for service in dc_cfg['services']:
        srv_cfg = deepcopy(dc_cfg['services'][service])

        result_service_name = services_map[service]
        new_service_dc_cfg['services'][result_service_name] = srv_cfg

        if 'depends_on' in srv_cfg:
            if isinstance(srv_cfg['depends_on'], list):
                new_deps = [
                    services_map[item] for item in srv_cfg['depends_on']
                ]
                new_service_dc_cfg['services'][result_service_name] = srv_cfg | {
                    'depends_on': new_deps
                }
            if isinstance(srv_cfg['depends_on'], dict):
                new_deps = {
                    services_map[service_name]: condition
                    for service_name, condition in srv_cfg['depends_on'].items()
                }
                new_service_dc_cfg['services'][result_service_name] = srv_cfg | {
                    'depends_on': new_deps
                }
    return new_service_dc_cfg


def patch_labels(dc_cfg: dict, release_id: str):
    new_service_dc_cfg = deepcopy(dc_cfg)
    for service in dc_cfg['services']:
        srv_cfg = deepcopy(dc_cfg['services'][service])
        if 'labels' not in srv_cfg:
            new_service_dc_cfg['services'][service]['labels'] = {}

        new_service_dc_cfg['services'][service]['labels'] = new_service_dc_cfg['services'][service]['labels'] | {
            'com.docker.maxwelld.release_id': release_id,
        }
    return new_service_dc_cfg


def patch_docker_compose_file_services(filename: Path,
                                       host_root: Path,
                                       services_environment_vars: Environment,
                                       network_name: str,
                                       # TODO network_name = [projectname]_default
                                       services_map: dict[str, str] | None,
                                       release_id: str = None,
                                       ) -> None:
    dc_cfg = read_dc_file(filename)

    dc_cfg = patch_network(dc_cfg, network_name=network_name)

    if release_id:
        dc_cfg = patch_labels(dc_cfg, release_id)

    if services_map:
        dc_cfg = patch_service_set(dc_cfg, services_map)  # todo use servcie_map

    if services_environment_vars:
        dc_cfg = patch_envs(dc_cfg, services_environment_vars)  # todo use servcie_map

    if services_map:
        dc_cfg = patch_services_names(dc_cfg, services_map)  # todo use servcie_map istead postfix

    dc_cfg = patch_services_volumes(dc_cfg, host_root)

    write_dc_file(filename, dc_cfg)


def parse_migration(migration: dict, service: str) -> Handler:
    if not isinstance(migration, dict):
        raise ServicesMigrationsError(
            f'In service: {service} migration line "{migration}" should be one of format:\n'
            f'  x-migration:\n'
            f'    - stage: command\n'
            f'    - stage: [[command], service]\n'
        )
    if len(migration) != 1:
        raise ServicesMigrationsError(
            f'In service: {service} migration "{migration}" should be nested like:\n'
            f'  x-migration:\n'
            f'    - stage: command\n'
            f'    - stage: [[command], service]\n'
        )

    if list(migration.keys())[0] not in EventStage.get_all_compose_stages():
        raise ServicesMigrationsError(
            f'In "{service}" migration: "{migration}" stage should only be one of '
            f'{EventStage.get_all_compose_stages()}'
        )

    for stage, command in migration.items():
        if isinstance(command, str):
            return Handler(
                EventStage.get_compose_stage(stage),
                cmd=command,
                executor=service,
            )

        if isinstance(command, list):
            assert len(command) >= 1, f"Migration {migration} command should have at least one command part"
            if isinstance(command[0], list):
                return Handler(
                    EventStage.get_compose_stage(stage),
                    cmd=' '.join(command[0]),
                    executor=command[1],
                )
            return Handler(
                EventStage.get_compose_stage(stage),
                cmd=' '.join(command),
                executor=service,
            )

        if isinstance(command, dict):
            raise ServicesMigrationsError(
                f'In service: {service} migration "{migration}" should be nested like:\n'
                f'  x-migration:\n'
                f'    - stage: command\n'
                f'    - stage: [[command], service]\n'
            )

        raise ServicesMigrationsError(f'migration {command[0]} should have one of:\n'
                                      f'x-migration:\n'
                                      f'  - stage: command\n'
                                      f'  - stage: [ command, cmd_part ]\n'
                                      '  - stage: [ [command, cmd_part], executor ]\n'
                                      '  - stage: { cmd: command, executor: executor_container_name}\n'
                                      f'  - stage:\n'
                                      f'      cmd: command\n'
                                      f'      executor: executor_container_name\n'
                                      f'formats')


def parse_migrations(migrations: list[dict], service: str) -> list[Handler]:
    if not isinstance(migrations, list):
        raise ServicesMigrationsError(f'In "{service}" migration:{migrations} should match format:\n'
                                      f'service:\n'
                                      f'  x-migration:\n'
                                      f'    - stage: command')

    result_hooks = []
    for migration in migrations:
        result_hooks += [parse_migration(migration, service)]
    return result_hooks


def extract_services_inline_migration(compose_files: list[str]) -> dict[str, list[str]]:
    migrations = collections.defaultdict(lambda: [])
    for filename in compose_files:
        dc_cfg = read_dc_file(filename)
        if 'services' in dc_cfg:
            for service in dc_cfg['services']:
                if service not in migrations:
                    migrations[service] = []

                if 'x-migration' in dc_cfg['services'][service]:
                    migrations[service] += parse_migrations(
                        dc_cfg['services'][service]['x-migration'],
                        service,
                    )

                if ('x-migrate' in dc_cfg['services'][service]) or ('x-migrations' in dc_cfg['services'][service]):
                    raise ServicesMigrationsError(
                        f'Service "{service}" have similar to migrations key, do u mean "x-migration" section?'
                    )

    return migrations


def make_env_compose_instance_files(env_config_instance: EnvInstanceConfig,
                                    compose_files: str,
                                    project_network_name: str,  # without "_default"
                                    host_project_root_directory,
                                    compose_files_path: Path,
                                    tmp_env_path: Path,
                                    release_id: str = None
                                    ) -> ComposeInstanceFiles:
    dst = tmp_env_path / env_config_instance.env_id

    dst.mkdir(parents=True, exist_ok=True)
    for file in dst.iterdir():
        if file.is_file():
            file.unlink()

    for file in compose_files.split(':'):
        src_file = compose_files_path / file
        dst_file = dst / file
        shutil.copy(src_file, dst_file)

        # TODO fill dc files with env from .envs files as default
        patch_docker_compose_file_services(
            dst_file,
            host_root=host_project_root_directory,
            services_environment_vars=env_config_instance.env,
            network_name=f'{project_network_name}_default',
            services_map=env_config_instance.env_services_map,
            release_id=release_id,
        )

    new_compose_files_list = get_new_instance_compose_files(compose_files, dst)

    inline_migrations = extract_services_inline_migration(new_compose_files_list.split(':'))

    return ComposeInstanceFiles(
        env_config_instance=env_config_instance,
        compose_files_source=compose_files,
        directory=dst,
        compose_files=new_compose_files_list,
        inline_migrations=inline_migrations,
    )


def get_compose_services(compose_files: str):
    services = []
    for filename in compose_files.split(':'):
        dc_cfg = read_dc_file(filename)
        if 'services' in dc_cfg:
            for service in dc_cfg['services'].keys():
                if service not in services:
                    services += [service]
    return services


def parse_docker_compose(file_paths):
    services_dict = {}

    for file_path in file_paths.split(':'):
        dc_cfg = read_dc_file(file_path)
        services = dc_cfg.get('services', {})
        for service_name, service_data in services.items():
            service_deps = set(services_dict.get(service_name, []))
            current_file_service_deps = set(service_data.get('depends_on', []))

            services_dict[service_name] = list(service_deps | current_file_service_deps)

    return services_dict


def topological_sort(services_dict):
    from collections import defaultdict, deque

    # Prepare the adjacency list
    graph = defaultdict(list)
    in_degree = defaultdict(int)

    for service, dependencies in services_dict.items():
        if service not in in_degree:
            in_degree[service] = 0
        for dependency in dependencies:
            graph[dependency].append(service)
            in_degree[service] += 1

    # Find all sources (services with 0 in-degree)
    zero_in_degree_queue = deque([service for service in services_dict if in_degree[service] == 0])
    topologically_sorted_services = []

    while zero_in_degree_queue:
        level_services = list(zero_in_degree_queue)
        for _ in range(len(zero_in_degree_queue)):
            node = zero_in_degree_queue.popleft()
            topologically_sorted_services.append(node)
            for neighbor in graph[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    zero_in_degree_queue.append(neighbor)

    if len(topologically_sorted_services) == len(services_dict):
        return topologically_sorted_services
    else:
        raise ValueError("A cycle detected in the services dependencies")


def group_by_levels(topologically_sorted_services, services_dict):
    service_levels = []
    service_to_level = {}

    # Assign level for each service
    for service in topologically_sorted_services:
        max_dependency_level = -1
        for dependency in services_dict[service]:
            if dependency in service_to_level:
                max_dependency_level = max(max_dependency_level, service_to_level[dependency])
        current_level = max_dependency_level + 1
        if current_level >= len(service_levels):
            service_levels.append([])
        service_levels[current_level].append(service)
        service_to_level[service] = current_level

    return service_levels


def get_compose_services_dependency_tree(compose_files: str):
    # Parse the docker-compose files to get service dependencies
    services_dict = parse_docker_compose(compose_files)

    # Perform topological sort to get an ordered list of services
    topologically_sorted_services = topological_sort(services_dict)

    # Group services by levels based on their dependencies
    service_levels = group_by_levels(topologically_sorted_services, services_dict)

    return service_levels



from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.version import get_version


async def healthcheck(request: Request) -> web.Response:
    return web.json_response({
        'status': 'ok',
        'version': get_version()
    })



from pathlib import Path
from typing import NamedTuple
from typing import Union

from maxwelld.env_description.env_types import Environment

EMPTY_ID = 'no_id'


class ComposeConfig(NamedTuple):
    compose_files: str
    parallel_env_limit: Union[int, None] = None


class EnvInstanceConfig(NamedTuple):
    env_source: Environment
    env_id: str
    env_services_map: dict[str, str]
    env: Environment


class ComposeInstanceFiles(NamedTuple):
    env_config_instance: EnvInstanceConfig
    compose_files_source: str
    directory: Path
    compose_files: str
    inline_migrations: dict = None



from typing import Iterator
from typing import List

from vedro.core import VirtualScenario

from maxwelld.env_description.env_types import DEFAULT_ENV
from maxwelld.env_description.env_types import Environment


def extract_scenario_config(scenario: VirtualScenario):
    scenario_env = DEFAULT_ENV
    if hasattr(scenario._orig_scenario, 'tags'):
        for tag in scenario._orig_scenario.tags:
            if isinstance(tag, Environment):
                scenario_env = str(tag)
    return scenario_env


def extract_scenarios_configs_set(scenarios: List[VirtualScenario] | Iterator[
    VirtualScenario]):
    needed_configs = set()
    for scenario in scenarios:
        needed_configs.add(extract_scenario_config(scenario))
    return needed_configs



'''
This module used by maxwelld vedro plugin.

All tasks for maxwelld service container transported via http transport.

Every command on service has it own endpoint with param set, described in maxwelld/server
'''



"""
Core if maxwelld service.

It stores all in flight environments, prerequisites for env, "up" and "down" sequences.

Base run sequence:
    - if requested environment (d-c files + service env params + name/purpose) exist
      -> return env_id
    - if requested environment doesnt ready
      -> config d-c files with service env params
      -> run environment
      -> run migrations for each service
      -> return env_id

All Maxwelld service commands made by sequence of such atomic docker-compose commands:
> docker-compose up -d %services%
> docker-compose down -d %services%
> docker-compose status
> docker-compose exec %service% %cmd%

Used docker-compose commands described in compose_interface
"""



from typing import TypedDict

from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.client.types import EnvironmentId
from maxwelld.core.service import MaxwellDemonServiceManager
from maxwelld.helpers.bytes_pickle import base64_pickled


class DcLogsRequestParams(TypedDict):
    env_id: EnvironmentId
    services: list[str]


class DcLogsResponseParams(TypedDict):
    logs: str


async def dc_logs(request: Request) -> web.Response:
    params: DcLogsRequestParams = await request.json()

    logs = await MaxwellDemonServiceManager().get().logs(
        env_id=params['env_id'],
        services=params['services'],
    )

    return web.json_response(DcLogsResponseParams(logs=base64_pickled(logs)), status=200)



from warnings import warn

from maxwelld.env_description.env_types import Service
from maxwelld.env_description.env_types import ServiceMode


def singleton(service: Service) -> Service:
    return Service(
        name=service.name,
        env=service.env,
        events_handlers=service.events_handlers,
        mode=ServiceMode.SINGLETON
    )


def off(service: Service) -> Service:
    warn('will be deprecated')
    return Service(
        name=service.name,
        env=service.env,
        events_handlers=service.events_handlers,
        mode=ServiceMode.OFF
    )



"""
This module used for the vedro plugin integration for Maxwell Demon

Enabled plugin (in vedro.cfg) catches stage before tests run (StartupEvent) and before each test run (ScenarioRunEvent).

On each stage it starts testing environment required by selected tests.

Plugin communicate with maxwelld service container via http client - maxwelld/client
"""



class CountdownCounterKeeper:
    def __init__(self, max_retries: int):
        self._count = 1
        self._max_retries = max_retries

    def tick(self):
        self._count += 1

    def is_done(self):
        return self._count > self._max_retries



import asyncio
import sys


async def process_output_till_done(process: asyncio.subprocess.Process, verbose) -> tuple[bytes, bytes]:
    stdout_lines = []
    stderr_lines = []

    async def read_stream(stream, callback, output_list):
        while True:
            line = await stream.readline()
            if line:
                if verbose:
                    callback(line)
                output_list.append(line)
            else:
                break

    tasks = [
        read_stream(process.stdout, lambda line: sys.stdout.buffer.write(b' > ' + line), stdout_lines),
        read_stream(process.stderr, lambda line: sys.stderr.buffer.write(b' > ' + line), stderr_lines)
    ]

    await asyncio.gather(*tasks)
    await process.wait()

    sys.stdout.flush()
    sys.stderr.flush()

    return b''.join(stdout_lines), b''.join(stderr_lines)



import warnings
from uuid import uuid4

from rich.text import Text

from maxwelld.client.types import EnvironmentId
from maxwelld.core.compose_data_types import ServicesComposeState
from maxwelld.core.compose_instance import ComposeInstanceManager
from maxwelld.core.compose_interface import ComposeShellInterface
from maxwelld.core.config import Config
from maxwelld.core.inflight_keeper import InflightKeeper
from maxwelld.core.sequence_run_types import EMPTY_ID
from maxwelld.core.utils.compose_instance_cfg import get_new_env_id
from maxwelld.core.utils.env_files import make_debug_bash_env
from maxwelld.env_description.env_types import Environment
from maxwelld.helpers.jobs_result import JobResult
from maxwelld.output.console import CONSOLE
from maxwelld.output.styles import Style


class MaxwellDemonService:
    def __init__(self,
                 config=Config,
                 compose_interface: type[ComposeShellInterface] = None,
                 inflight_keeper: type[InflightKeeper] = None,
                 compose_instance_maker: type[ComposeInstanceManager] = None):
        # assert shutil.which("docker"), 'Docker not installed'
        # assert shutil.which("docker-compose"), 'Docker-compose not installed'

        # assert os.environ.get('COMPOSE_PROJECT_NAME'), \
        #     'COMPOSE_PROJECT_NAME env should be set'
        # assert os.environ.get('NON_STOP_CONTAINERS'), \
        #     'NON_STOP_CONTAINERS env should be set'
        # assert os.environ.get('HOST_PROJECT_ROOT_DIRECTORY'), \
        #     'HOST_PROJECT_ROOT_DIRECTORY env should be set'

        # TODO print service paths for *.yml files, project root and etc.
        cfg = config()
        self._project = cfg.project
        self._non_stop_containers = cfg.non_stop_containers
        self.tmp_envs_path = cfg.tmp_envs_path
        self.compose_files_path = cfg.compose_files_path
        self.in_docker_project_root_path = cfg.in_docker_project_root_path
        self.env_file_name = cfg.env_file_name
        self.env_file_path = cfg.env_file_path
        self.host_project_root_directory = cfg.host_project_root_directory
        self.env_tmp_directory = cfg.env_tmp_directory
        self.host_env_tmp_directory = cfg.host_env_tmp_directory

        self._compose_interface = ComposeShellInterface
        if compose_interface is not None:
            self._compose_interface = compose_interface

        self._compose_instance_manager = ComposeInstanceManager(
            project=self._project,
            compose_interface=self._compose_interface,
            except_containers=self._non_stop_containers,
            compose_files_path=cfg.compose_files_path,
            default_compose_files=cfg.default_compose_files,
            in_docker_project_root=self.in_docker_project_root_path,
            host_project_root_directory=self.host_project_root_directory,
            tmp_envs_path=self.tmp_envs_path,
        )

        if compose_instance_maker is not None:
            self._compose_instance_manager = compose_instance_maker(
                project=self._project,
                compose_interface=self._compose_interface,
                except_containers=self._non_stop_containers,
                compose_files_path=cfg.compose_files_path,
                default_compose_files=cfg.default_compose_files,
                in_docker_project_root=self.in_docker_project_root_path,
                host_project_root_directory=self.host_project_root_directory,
                tmp_envs_path=self.tmp_envs_path,
            )

        if inflight_keeper is None:
            self._inflight_keeper = InflightKeeper(self.tmp_envs_path, self.env_file_name)

        self._inflight_keeper.cleanup_in_flight()

    def _unpack_services_env_template_params(self, env: Environment):
        return {service: env[service].env for service in env}

    async def up_compose(self, name: str, config_template: Environment, compose_files: str, isolation=None,
                         parallelism_limit=None, verbose=False, force_restart=False) -> tuple[EnvironmentId, bool]:
        warnings.warn('Deprecated, use up_or_get_existing instead')
        return await self.up_or_get_existing(
            name, config_template, compose_files, isolation, parallelism_limit, verbose, force_restart
        )

    async def get_existing(self, name: str, config_template: Environment | None, compose_files: str | None):
        existing_inflight_env = self._inflight_keeper.get_existing_inflight_env(
            name, config_template, compose_files
        )
        if existing_inflight_env:
            CONSOLE.print(f'Existing env for {name}: {existing_inflight_env.env_id}. Access: '
                          f'> cd {self.host_project_root_directory} && '
                          f'source ./env-tmp/{existing_inflight_env.env_id}/.env')
            return existing_inflight_env.env_id
        return None

    async def up_or_get_existing(
        self, name: str, config_template: Environment | None, compose_files: str | None, isolation=None,
        parallelism_limit=None,
        verbose=False, force_restart: bool = False,
        release_id: str = None,
    ) -> tuple[EnvironmentId, bool]:

        existing_inflight_env_id = await self.get_existing(name, config_template, compose_files)
        # TODO check all services up
        if existing_inflight_env_id and not force_restart:
            return existing_inflight_env_id, False

        CONSOLE.print(
            Text('Starting new environment: ', style=Style.info)
            .append(Text(name, style=Style.mark))
        )

        new_env_id = get_new_env_id()
        if parallelism_limit == 1:
            CONSOLE.print(f'Using default service names with {parallelism_limit=}')
            new_env_id = EMPTY_ID

        target_compose_instance = self._compose_instance_manager.make(
            new_env_id,
            compose_files=compose_files,
            config_template=config_template,
            release_id=release_id,
        )
        if parallelism_limit == 1:
            # check if limit 1 - existing already not fit - down all current inflight
            to_down = self._compose_instance_manager.get_envs()
            for instance in to_down:
                await self._compose_instance_manager.down_env(instance)
                self._inflight_keeper.cleanup_in_flight()
                # TODO check if > 1
                #          check current {name} is runnig?
                #              runnig -> current {name} to down list
                #              check curren - 1 > limit
                #                   grab to down some of limit - (current - 1)

        await target_compose_instance.cleanup()
        await target_compose_instance.run()
        target_compose_instance_files = target_compose_instance.compose_instance_files

        make_debug_bash_env(target_compose_instance_files, self.host_env_tmp_directory)
        CONSOLE.print(f'Docker-compose access: > cd {self.host_project_root_directory} && '
                      f'source ./env-tmp/{new_env_id}/.env')

        # TODO should be transactional with file
        CONSOLE.print(Text(f'New environment for {name} started'))
        if verbose:
            CONSOLE.print(
                f'Config params: {self._unpack_services_env_template_params(target_compose_instance.config().env)}'
            )
        CONSOLE.print(
            Text(f'Docker-compose access: > ', style=Style.info)
            .append(Text(
                f'cd {self.host_project_root_directory} && source ./env-tmp/{new_env_id}/.env',
                style=Style.mark_neutral))
        )

        self._inflight_keeper.update_inflight_envs(
            name,
            config_template=config_template,
            compose_files=target_compose_instance_files.compose_files,
            env_id=new_env_id,
        )

        return new_env_id, True

    def env(self, env_id: str) -> Environment:
        env_instance_config = self._inflight_keeper.get_existing_inflight_env_by_id(env_id)
        return env_instance_config.env

    async def status(self, env_id: str) -> ServicesComposeState:
        env_compose_files = self._inflight_keeper.get_existing_inflight_env_compose_files(env_id)
        services_status = await self._compose_interface(
            compose_files=env_compose_files,
            in_docker_project_root=self.in_docker_project_root_path
        ).dc_state()
        assert isinstance(services_status, ServicesComposeState), "Can't execute docker-compose ps"
        return services_status

    async def exec(self, env_id: str, container: str, command: str):
        log_file = f'{str(uuid4())}.log'
        env_compose_files = self._inflight_keeper.get_existing_inflight_env_compose_files(env_id)
        compose_interface = self._compose_interface(
            compose_files=env_compose_files,
            in_docker_project_root=self.in_docker_project_root_path
        )

        await compose_interface.dc_exec(container, f'sh -c \'{command} > /tmp/{log_file}\'')

        job_result, stdout, stderr = await compose_interface.dc_exec(container, f'cat /tmp/{log_file}')
        if job_result != JobResult.GOOD:
            ...

        return stdout

    async def logs(self, env_id: str, services: list[str]) -> dict[str, bytes]:
        env_compose_files = self._inflight_keeper.get_existing_inflight_env_compose_files(env_id)
        compose_interface = self._compose_interface(
            compose_files=env_compose_files,
            in_docker_project_root=self.in_docker_project_root_path
        )

        logs = {}
        for service in services:
            job_result, log = await compose_interface.dc_logs([service])
            logs[service] = log
            if job_result != JobResult.GOOD:
                ...

        return logs


class MaxwellDemonServiceManager:
    maxwell_demon_service = None

    def __init__(self):
        if MaxwellDemonServiceManager.maxwell_demon_service is None:
            MaxwellDemonServiceManager.maxwell_demon_service = MaxwellDemonService()

    def get(self) -> MaxwellDemonService:
        return MaxwellDemonServiceManager.maxwell_demon_service



import sys
from enum import Enum
from enum import auto
from functools import partial
from functools import reduce
from sys import exit
from typing import Type
from typing import Union

import vedro.events
from rich.text import Text
from vedro.core import ConfigType
from vedro.core import Dispatcher
from vedro.core import Plugin
from vedro.core import PluginConfig
from vedro.events import ArgParseEvent
from vedro.events import ArgParsedEvent
from vedro.events import ConfigLoadedEvent
from vedro.events import ScenarioRunEvent
from vedro.events import StartupEvent

from maxwelld.client.maxwell_client import MaxwellDemonClient
from maxwelld.core.sequence_run_types import ComposeConfig
from maxwelld.env_description.env_types import Environments
from maxwelld.helpers.jobs_result import JobResult
from maxwelld.output.console import CONSOLE
from maxwelld.output.styles import Style
from maxwelld.vedro_plugin.env_setter import setup_env_for_tests
from maxwelld.vedro_plugin.logger import WaitVerbosity
from maxwelld.vedro_plugin.scenario_ordering import EnvTagsOrderer
from maxwelld.vedro_plugin.scenario_tag_processing import extract_scenario_config
from maxwelld.vedro_plugin.scenario_tag_processing import extract_scenarios_configs_set
from maxwelld.vedro_plugin.state_waiting import wait_all_services_up

DEFAULT_COMPOSE = 'default'


class Stage(Enum):
    INIT = auto()
    PRE_TEST = auto()


class VedroMaxwellPlugin(Plugin):
    def __init__(self, config: Type["VedroMaxwell"]) -> None:
        super().__init__(config)
        self._enabled = config.enabled
        self._envs: Environments = config.envs
        self._maxwell_demon = MaxwellDemonClient(host="http://127.0.0.1")
        if config.maxwell_demon_client:
            self._maxwell_demon = config.maxwell_demon_client
        self._list_envs = None

        self._compose_configs: dict[str, ComposeConfig] = config.compose_cfgs
        assert DEFAULT_COMPOSE in self._compose_configs, \
            'Need to set up at least {DEFAULT_COMPOSE: ComposeConfig(...)} config'
        self._compose_choice_name: str = DEFAULT_COMPOSE
        self._compose_choice: Union[ComposeConfig, None] \
            = self._compose_configs[self._compose_choice_name]
        self._force_env_name: Union[str, None] = None
        self._chosen_config_name_postfix: str = ''
        self._checked_envs = []
        self._wait_all_service_func = config.wait_all_service_func
        self._force_restart = False

        self._lats_env_name_started = None
        self._lats_env_id_started = None

        self._reported_full = False
    def _print_running_config(self):
        CONSOLE.print(
            Text('Running ', style=Style.info)
            .append(Text(self._compose_choice_name, style=Style.mark))
            .append(Text(' compose config: '))
            .append(Text(str(self._compose_choice), style=Style.mark))
        )
        if self._force_env_name:
            CONSOLE.print(f'Overriding configuration for tests: {self._force_env_name}')

    def subscribe(self, dispatcher: Dispatcher) -> None:
        if not self._enabled:
            return

        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
            .listen(vedro.events.ArgParseEvent, self.handle_arg_parse) \
            .listen(vedro.events.ArgParsedEvent, self.handle_arg_parsed) \
            .listen(vedro.events.StartupEvent, self.handle_scenarios) \
            .listen(vedro.events.ScenarioRunEvent, self.handle_setup_test_config)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config: ConfigType = event.config

    async def wait_env_ready(self, env_id, verbose: WaitVerbosity) -> None:
        environment = await self._maxwell_demon.env(env_id)

        checker = self._wait_all_service_func.make_checker()

        up_result = await checker(
            get_services_state=partial(self._maxwell_demon.status, env_id=env_id),
            services=environment.get_services(),
            verbose=verbose,
        )
        assert up_result != JobResult.BAD, f"Can't done up environment"

    async def up_env(self, env_name, stage):
        if (self._verbose and not self._reported_full) or stage == Stage.INIT:
            CONSOLE.print(
                Text('Starting ', style=Style.regular)
                .append(Text(env_name, style=Style.mark))
                .append(Text(' services for tests ...', style=Style.regular))
            )

        env = getattr(self._envs, env_name)
        started_env_id = await self._maxwell_demon.up(
            name=env_name + self._chosen_config_name_postfix,
            config_template=env,
            compose_files=self._compose_choice.compose_files,
            parallelism_limit=self._compose_choice.parallel_env_limit,
            force_restart=self._force_restart
        )

        verbose = WaitVerbosity.COMPACT
        if self._lats_env_id_started == started_env_id:
            verbose = WaitVerbosity.ON_ERROR
        if self._verbose and not self._reported_full:
            verbose = WaitVerbosity.FULL

        await self.wait_env_ready(
            env_id=started_env_id,
            verbose=verbose
        )

        self._force_restart = False
        self._reported_full = True
        self._lats_env_id_started = started_env_id
        return started_env_id

    async def handle_scenarios(self, event: StartupEvent) -> None:
        needed_configs = extract_scenarios_configs_set(event.scheduler.scheduled)
        if not needed_configs:
            return

        CONSOLE.print(
            Text(f'Tests requests configs: ') + reduce(
                lambda a, b: a + Text(', ', style=Style.regular) + b, [
                    Text(f'{cfg}', style=Style.mark) for cfg in needed_configs
                ]
            )
        )

        if self._force_env_name:
            config_env_name = self._force_env_name
            CONSOLE.print(f'Overriding tests config:{needed_configs} '
                          f'by --env={self._force_env_name}')
            needed_configs = {config_env_name}

        if self._list_envs:
            sys.exit()

        if (
                self._compose_choice.parallel_env_limit
                and (len(needed_configs) > self._compose_choice.parallel_env_limit)
        ):
            self._global_config.Registry.ScenarioOrderer.register(EnvTagsOrderer, self)

        if (
                (self._compose_choice.parallel_env_limit is None)
                or (self._compose_choice.parallel_env_limit == len(needed_configs))
        ):
            for cfg_name in list(needed_configs):
                await self._maxwell_demon.healthcheck()
                await self.up_env(cfg_name, Stage.INIT)

    async def handle_setup_test_config(self, event: ScenarioRunEvent):
        config_env_name = extract_scenario_config(event.scenario_result.scenario)

        if self._force_env_name:
            if self._verbose:
                CONSOLE.print(f'Overriding tests config: '
                              f'{config_env_name} by --md-env={self._force_env_name}')
            config_env_name = self._force_env_name

        env_id = await self.up_env(config_env_name, Stage.PRE_TEST)

        environment = await self._maxwell_demon.env(env_id)
        setup_env_for_tests(environment)

    def handle_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Maxwell Demon")
        group.add_argument("--md-list-envs",
                           action='store_true',
                           help="List possible enviroments")

        for choice_name, config in self._compose_configs.items():
            default_text = '[set by default]' if choice_name == DEFAULT_COMPOSE else ''
            group.add_argument(f"--md-{choice_name}",
                               action='store_true',
                               help=f"Choose compose config {default_text}: {config}")

        group.add_argument("--md-fr",
                           action='store_true',
                           help="Force restart env")

        group.add_argument("--md-list-services",
                           action='store_true',
                           help="List possible enviroments")

        group.add_argument("--md-env",
                           type=str,
                           choices=list(self._envs.list_all()),
                           help="Up choosen enviroment")

        group.add_argument("--md-v",
                           action='store_true',
                           help="List possible enviroments")

        group.add_argument("--md-parallel-env-limit",
                           type=int,
                           help="Max of parallel running envs, unused will be killed when extra "
                                "one needs")

    def handle_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._parallel_env_limit = event.args.md_parallel_env_limit
        self._verbose = event.args.md_v

        for choice_name, config in self._compose_configs.items():
            if getattr(event.args, f'md_{choice_name}'):
                self._compose_choice_name = choice_name
                self._compose_choice = config
                self._parallel_env_limit = config.parallel_env_limit
                if choice_name != DEFAULT_COMPOSE:
                    self._chosen_config_name_postfix = f'_{choice_name}'

        self._list_envs = event.args.md_list_envs
        if self._list_envs:
            self._maxwell_demon.list_current_in_flight_envs(self._envs.list_all())
            exit(0)

        if event.args.md_list_services:
            self._maxwell_demon.list_services()
            exit(0)

        if event.args.md_env:
            self._force_env_name = event.args.md_env

        if event.args.md_fr:
            self._force_restart = event.args.md_fr

        self._print_running_config()


class VedroMaxwell(PluginConfig):
    plugin = VedroMaxwellPlugin

    # Enables plugin
    enabled = False

    # Maxwell environments
    envs: Environments = None

    # Maxwell Demon To control the world
    maxwell_demon_client: MaxwellDemonClient = None

    # ComposeConfig set of compose files and defaulr parallelism restrictions
    compose_cfgs: dict[str, ComposeConfig] = None

    # services waiter
    wait_all_service_func = wait_all_services_up()



from typing import List

from vedro.core import ScenarioOrderer
from vedro.core import VirtualScenario

from maxwelld.vedro_plugin.scenario_tag_processing import extract_scenario_config


class EnvTagsOrderer(ScenarioOrderer):
    async def sort(self, scenarios: List[VirtualScenario]) -> List[VirtualScenario]:
        copied = scenarios[:]

        return sorted(
            copied,
            key=lambda x: extract_scenario_config(x)
        )



from copy import deepcopy
from enum import Enum
from enum import auto
from typing import TypeVar

T = TypeVar('T')


class ServicesState(Enum):
    FIRST_STATE = auto()
    DEFAULT_STATE = auto()


class StateKeeper:
    def __init__(self, state: T | ServicesState = ServicesState.FIRST_STATE):
        self._state: T | ServicesState = deepcopy(state)

    def in_state(self, new_state: T | ServicesState):
        return self._state == new_state

    def not_in_state(self, new_state: T | ServicesState):
        return self._state != new_state

    def update_state(self, new_state: T | ServicesState):
        self._state = deepcopy(new_state)



import asyncio
import traceback
from typing import TypedDict

from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.client.types import EnvironmentId
from maxwelld.core.service import MaxwellDemonServiceManager
from maxwelld.errors.up import ServicesUpError
from maxwelld.helpers.bytes_pickle import debase64_pickled

UP_LOCK = asyncio.Lock()


class DcUpRequestParams(TypedDict):
    name: str | None
    config_template: str | None
    compose_files: str | None
    isolation: bool | None
    parallelism_limit: int | None
    force_restart: bool | None


DC_UP_DEFAULTS = {
    'name': 'DEFAULT_FULL',
    'config_template': None,
    'compose_files': None,
    'isolation': False,
    'parallelism_limit': 1,
    'force_restart': False,
}


class UpResponseParams(TypedDict):
    env_id: EnvironmentId


class UpErrorResponseParams(TypedDict):
    error: str


async def dc_up(request: Request) -> web.Response:
    params: DcUpRequestParams = DC_UP_DEFAULTS | await request.json()
    config_template = debase64_pickled(params['config_template']) if params['config_template'] else None
    release_id = request.headers.get('x-release-id', None)

    # TODO move to up_or_get_existing
    # TODO kill existing composes??
    try:
        env_id, is_new = await MaxwellDemonServiceManager().get().up_or_get_existing(
            name=params['name'],
            config_template=config_template,
            compose_files=params['compose_files'],
            isolation=params['isolation'],
            parallelism_limit=params['parallelism_limit'],
            force_restart=params['force_restart'],
            release_id=release_id
        )
    except ServicesUpError as e:
        return web.json_response(UpErrorResponseParams(error=e.message), status=422)
    except AssertionError as e:
        return web.json_response(UpErrorResponseParams(error=f'Somthing went wrong:\n{str(e)}'), status=422)
    except BaseException as e:
        return web.json_response(UpErrorResponseParams(error=f'Somthing went terribly wrong:\n{str(e)}\n\n{traceback.format_exc()}'), status=422)
    return web.json_response(UpResponseParams(env_id=env_id), status=200)



from enum import Enum
from enum import auto

from rich.text import Text

from maxwelld.output.styles import Style


class WaitVerbosity(Enum):
    ON_ERROR = auto()
    COMPACT = auto()
    FULL = auto()


class Logger:
    def __init__(self, console):
        self._console = console
        self._log: Text | None = None

    def log(self, text: Text):
        if self._log is None:
            self._log = text
            return
        self._log.append(Text('\n', style=Style.regular).append(text))

    def flush(self):
        if self._log:
            self._console.print(self._log + Text(' ', style=Style.regular))



from pathlib import Path
from uuid import uuid4

from maxwelld.client.types import EnvironmentId
from maxwelld.core.sequence_run_types import EMPTY_ID
from maxwelld.core.sequence_run_types import EnvInstanceConfig
from maxwelld.env_description.env_types import Env
from maxwelld.env_description.env_types import Environment
from maxwelld.env_description.env_types import Service
from maxwelld.env_description.env_types import ServiceMode


def make_env_service_name(service, env_id):
    if env_id == EMPTY_ID:
        return service
    return f'{service}-{env_id}'


def sub_env_id(env: Env, services_for_env) -> Env:
    result_env = Env()
    for k, v in env.items():
        # TODO split environments by services
        try:
            res = v.format(**services_for_env)
            result_env.update({k: res})
        except KeyError:
            # temporarily suppressed
            # warn(f'cant substitute key: {k} for value: {v}, with {services_for_env}')
            ...

    return result_env


def prepare_services_env(env: Environment, services_map: dict) -> Environment:
    updated_services = []
    for service_name in env:
        updated_services += [
            Service(
                name=env[service_name].name,
                env=sub_env_id(env[service_name].env, services_map),
                events_handlers=env[service_name].events_handlers,
                mode=env[service_name].mode,
            )
        ]
    new_env = Environment(str(env), *updated_services)
    return new_env


def get_new_env_id() -> EnvironmentId:
    env_id = str(uuid4())[:4]
    return env_id


def get_service_map(env: Environment, new_env_id: str):
    return {
        service_name: make_env_service_name(service_name, new_env_id)
        for service_name, service in env.get_services().items() if service.mode != ServiceMode.OFF
    }


def make_env_instance_config(env_template: Environment | None, env_id) -> EnvInstanceConfig:
    services_map = None
    if env_template:
        services_map = get_service_map(env_template, env_id)

    env = Environment('DEFAULT_FULL')
    if env_template:
        env = prepare_services_env(env_template, services_map)

    return EnvInstanceConfig(
        env_source=env_template,
        env_id=env_id,
        env_services_map=services_map,
        env=env
    )


def get_new_instance_compose_files(compose_files: str, env_directory: Path) -> str:
    return ':'.join(
        [
            str(env_directory / compose_file)
            for compose_file in compose_files.split(':')
        ]
    )



"""
This module used for http server for Maxwell Demon container

Client for this server - maxwelld/client

Each command contains params set described in it's handler (for example up_compose: UpRequestParams)
Param set is used by client for serializing into json; and by server for deserializing same param set from json.
Same happens for response params (for example http_get_status: StatusResponseParams).

For incoming request it searches for initialized maxwelld service instance (core/service) and runs 'up', 'status'
or any other commands.
"""



import os
import sys
from pathlib import Path

from rich.text import Text
from yaml.parser import ParserError

from maxwelld.core.compose_interface import ComposeShellInterface
from maxwelld.core.config import Config
from maxwelld.core.sequence_run_types import ComposeInstanceFiles
from maxwelld.core.sequence_run_types import EnvInstanceConfig
from maxwelld.core.utils.compose_files import get_compose_services
from maxwelld.core.utils.compose_files import get_compose_services_dependency_tree
from maxwelld.core.utils.compose_files import make_env_compose_instance_files
from maxwelld.core.utils.compose_instance_cfg import get_new_instance_compose_files
from maxwelld.core.utils.compose_instance_cfg import make_env_instance_config
from maxwelld.env_description.env_types import Environment
from maxwelld.env_description.env_types import EventStage
from maxwelld.errors.up import ServicesUpError
from maxwelld.helpers.jobs_result import JobResult
from maxwelld.output.console import CONSOLE
from maxwelld.output.styles import Style
from maxwelld.vedro_plugin.logger import WaitVerbosity
from maxwelld.vedro_plugin.state_waiting import wait_all_services_up

INFLIGHT = 'inflight'


class ComposeInstance:
    def __init__(self,
                 project: str,
                 new_env_id: str,
                 compose_interface: type[ComposeShellInterface],
                 compose_files: str,
                 compose_files_path: Path,
                 config_template: Environment | None,
                 in_docker_project_root: Path,
                 host_project_root_directory: Path,
                 except_containers: list[str],
                 tmp_envs_path: Path,
                 execution_envs: dict = None,
                 release_id: str = None):
        self.compose_files = compose_files
        self.compose_files_path = compose_files_path
        self.in_docker_project_root = in_docker_project_root
        self.host_project_root_directory = host_project_root_directory
        self.except_containers = except_containers
        self.tmp_envs_path = tmp_envs_path

        self.project = project
        if execution_envs is None:
            self.execution_envs = dict(os.environ)

        self.new_env_id = new_env_id
        self.config_template = config_template
        self._env_instance_config = None
        self.compose_interface = compose_interface

        self.release_id: str = release_id

        self.compose_instance_files: ComposeInstanceFiles = None
        for file in self.compose_files.split(':'):
            assert (file := Path(self.compose_files_path / file)).exists(), f'File {file} doesnt exist'

    async def config(self) -> EnvInstanceConfig:
        if self._env_instance_config is None:
            self._env_instance_config = make_env_instance_config(
                env_template=self.config_template,
                env_id=self.new_env_id
            )
        return self._env_instance_config

    async def generate_config_files(self) -> ComposeInstanceFiles:
        assert self.new_env_id != INFLIGHT, 'somehow u try to regenerate files for inflight env'
        compose_instance_files = make_env_compose_instance_files(
            await self.config(),
            self.compose_files,
            project_network_name=self.project,
            host_project_root_directory=self.host_project_root_directory,
            compose_files_path=self.compose_files_path,
            tmp_env_path=self.tmp_envs_path,
            release_id=self.release_id,
        )
        # TODO uneven compose_executor initialization!! but compose_interface compose_files-dependent
        self.compose_executor = self.compose_interface(
            compose_files=compose_instance_files.compose_files,
            in_docker_project_root=self.in_docker_project_root,
        )
        return compose_instance_files

    async def run_migration(self, stages, services, env_config_instance, migrations):
        for service in env_config_instance.env:
            sys.stdout.flush()
            if service not in services:
                continue

            for handler in migrations[service]:
                if handler.stage not in stages:
                    continue

                # TODO fix service map if default env
                target_service = env_config_instance.env_services_map[handler.executor or service]

                substituted_cmd = handler.cmd.format(**env_config_instance.env_services_map)
                migrate_result, stdout, stderr = await self.compose_executor.dc_exec(
                    target_service, substituted_cmd
                )
                if migrate_result != JobResult.GOOD:
                    services_status = await self.compose_executor.dc_state()
                    raise ServicesUpError(f"Can't migrate service {target_service}, with {substituted_cmd}"
                                          f"\n{stdout=}\n{stderr=}"
                                          f"\nUp logs:\n {await self.logs(self.except_containers)}"
                                          f"\nServices logs:\n {await self.logs(services)}"
                                          f"\nServices status:\n {services_status.as_rich_text()}") from None

    async def run_services_pack(self, services: list[str], migrations):

        for container in self.except_containers:
            if container in services:
                services.remove(container)
        print(f'Starting services pack: {services}; except: {self.except_containers}')

        status_result = await self.compose_executor.dc_state()
        assert status_result != JobResult.BAD, f"Can't get first status for services {services}"

        await self.run_migration(
            [EventStage.BEFORE_SERVICE_START],
            services,
            self.compose_instance_files.env_config_instance,
            migrations
        )

        up_result = await self.compose_executor.dc_up(services)
        assert up_result == JobResult.GOOD, (f"Can't up services {services}\nServices logs:\n "
                                             f"{await self.logs()}")
        services_status = await self.compose_executor.dc_state()

        # !!!!!! up process asynchronous
        # !!!!!! check services started before migrations

        await self.run_migration(
            [EventStage.AFTER_SERVICE_START],
            services,
            self.compose_instance_files.env_config_instance,
            migrations
        )

        checker = wait_all_services_up(
            attempts=Config().service_up_check_attempts,
            delay_s=Config().service_up_check_delay
        ).make_checker()
        check_up_result = await checker(
            get_services_state=self.compose_executor.dc_state,
            services=services,
            verbose=WaitVerbosity.COMPACT
        )
        if check_up_result != JobResult.GOOD:
            services_status = await self.compose_executor.dc_state()
            raise ServicesUpError(f"Can't up services {services} for "
                                  f"{Config().service_up_check_attempts*Config().service_up_check_delay}s"
                                  f"\nUp logs:\n {await self.logs(self.except_containers)}"
                                  f"\nServices logs:\n {await self.logs(services)}"
                                  f"\nServices status:\n {services_status.as_rich_text()}") from None

        await self.run_migration(
            [EventStage.AFTER_SERVICE_HEALTHY],
            services,
            self.compose_instance_files.env_config_instance,
            migrations
        )

    async def cleanup(self):
        ...

    async def run(self):
        self.compose_instance_files = await self.generate_config_files()

        services_tiers = get_compose_services_dependency_tree(self.compose_instance_files.compose_files)
        CONSOLE.print(
            Text('Starting services: ', style=Style.info)
            .append(Text(str(services_tiers), style=Style.good))
        )

        migrations = {}
        for service in self.compose_instance_files.env_config_instance.env:
            if service not in migrations:
                migrations[service] = []
            migrations[service] += self.compose_instance_files.env_config_instance.env[service].events_handlers
            migrations[service] += self.compose_instance_files.inline_migrations[service]

        all_services = [
            service
            for service_tier_pack in services_tiers
            for service in service_tier_pack
        ]

        await self.run_migration(
            [EventStage.BEFORE_ALL],
            all_services,
            self.compose_instance_files.env_config_instance,
            migrations
        )

        for service_tier_pack in services_tiers:
            await self.run_services_pack(service_tier_pack, migrations)

        await self.run_migration(
            [EventStage.AFTER_ALL],
            all_services,
            self.compose_instance_files.env_config_instance,
            migrations
        )

    async def logs(self, services=None) -> str:
        job_result, log = await self.compose_executor.dc_logs(services, logs_param='')
        return log.decode('utf-8') if job_result == JobResult.GOOD else ''


class ComposeInstanceManager:
    def __init__(self, project: str, compose_interface: type[ComposeShellInterface], except_containers: list[str],
                 compose_files_path: Path, default_compose_files: str, in_docker_project_root: Path,
                 host_project_root_directory: Path,
                 tmp_envs_path: Path):
        self.project = project
        self.compose_interface = compose_interface
        self.except_containers = except_containers
        self.compose_files_path = compose_files_path
        self.default_compose_files = default_compose_files
        self.in_docker_project_root = in_docker_project_root
        self.host_project_root_directory = host_project_root_directory
        self.tmp_envs_path = tmp_envs_path

    def make(self, new_env_id: str, compose_files: str | None, config_template: Environment, release_id: str = None):
        return ComposeInstance(
            project=self.project,
            compose_interface=self.compose_interface,
            new_env_id=new_env_id,
            compose_files=compose_files if compose_files else self.default_compose_files,
            compose_files_path=self.compose_files_path,
            config_template=config_template,
            in_docker_project_root=self.in_docker_project_root,
            host_project_root_directory=self.host_project_root_directory,
            except_containers=self.except_containers,
            tmp_envs_path=self.tmp_envs_path,
            release_id=release_id,
        )

    def from_compose_instance_files(self, compose_files: str | None, config_template: Environment, ):
        return ComposeInstance(
            project=self.project,
            compose_interface=self.compose_interface,
            new_env_id=INFLIGHT,
            compose_files=compose_files if compose_files else self.default_compose_files,
            compose_files_path=self.compose_files_path,
            config_template=config_template,
            in_docker_project_root=self.in_docker_project_root,
            host_project_root_directory=self.host_project_root_directory,
            except_containers=self.except_containers,
            tmp_envs_path=self.tmp_envs_path,
        )

    def get_envs(self) -> list[str]:
        dirpath, dirnames, filenames = next(os.walk(self.tmp_envs_path))
        return dirnames

    async def down_env(self, env_id):
        dirpath, dirnames, filenames = next(os.walk(self.tmp_envs_path / env_id))
        if '.env' in filenames:
            filenames.remove('.env')
        docker_files = get_new_instance_compose_files(':'.join(filenames), self.tmp_envs_path / env_id)

        try:
            services = get_compose_services(docker_files)
        except ParserError:
            return []
        print(f'Down services: {services}, except {self.except_containers}')
        for container in self.except_containers:
            if container in services:
                services.remove(container)

        compose_executor = self.compose_interface(
            compose_files=docker_files,
            in_docker_project_root=self.in_docker_project_root,
        )
        down_result = await compose_executor.dc_down(services)
        assert down_result == JobResult.GOOD, (f"Can't down services {services}")



from functools import partial
from typing import Awaitable
from typing import Callable
from typing import List

from rich.text import Text
from rtry import retry

from maxwelld.core.compose_data_types import ComposeHealth
from maxwelld.core.compose_data_types import ComposeState
from maxwelld.core.compose_data_types import ServiceComposeState
from maxwelld.core.compose_data_types import ServicesComposeState
from maxwelld.helpers.countdown_counter import CountdownCounterKeeper
from maxwelld.helpers.jobs_result import JobResult
from maxwelld.helpers.state_keeper import ServicesState
from maxwelld.helpers.state_keeper import StateKeeper
from maxwelld.output.console import CONSOLE
from maxwelld.output.styles import Style
from maxwelld.vedro_plugin.logger import Logger
from maxwelld.vedro_plugin.logger import WaitVerbosity


def is_service_not_running_or_not_healthy(service_state: ServiceComposeState) -> bool:
    return (service_state.state != ComposeState.RUNNING
            or service_state.health not in (ComposeHealth.EMPTY, ComposeHealth.HEALTHY))


async def check_all_services_up(
    get_services_state: Callable[[], Awaitable[ServicesComposeState]],
    services: List[str],
    counter_keeper: CountdownCounterKeeper,
    state_keeper: StateKeeper,
    verbose: WaitVerbosity = WaitVerbosity.FULL,
) -> JobResult:
    output_style = Style()
    logger = Logger(CONSOLE)

    if state_keeper.in_state(ServicesState.FIRST_STATE):
        logger.log(Text('Starting services check up', style=output_style.info))
        state_keeper.update_state(ServicesState.DEFAULT_STATE)

    services_state = await get_services_state()
    is_all_up = (all([
        service.state == ComposeState.RUNNING
        for service in services_state if service.name in services
    ]) and all([
        service.health in (ComposeHealth.EMPTY, ComposeHealth.HEALTHY)
        for service in services_state if service.name in services
    ]))

    if is_all_up:
        if verbose == WaitVerbosity.COMPACT:
            logger.log(Text(f' ✔ All services up\n', style=output_style.mark_neutral))
            logger.flush()
        if verbose == WaitVerbosity.FULL:
            logger.log(Text(f' ✔ All services up:', style=output_style.mark_neutral))
            logger.log(services_state.as_rich_text(style=output_style))
            logger.flush()
        return JobResult.GOOD

    counter_keeper.tick()
    if counter_keeper.is_done():
        logger.log(Text(' ✗ Stop retries. Still not ready services:', style=output_style.bad))
        logger.log(services_state.as_rich_text(style=output_style))
        logger.flush()
        return JobResult.BAD

    if state_keeper.not_in_state(services_state):
        logger.log(Text(f' ✗ Still not ready services:', style=output_style.bad))
        logger.log(services_state.as_rich_text(
            filter=is_service_not_running_or_not_healthy,
            style=output_style
        ))
        logger.flush()
        state_keeper.update_state(services_state)

    return JobResult.BAD


class WaitAllServicesUp:
    def __init__(self, attempts: int = 100, delay_s: int = 3):
        self._attempts = attempts
        self._delay_s = delay_s

    def make_checker(self) -> Callable:
        return partial(
            retry(
                attempts=self._attempts,
                delay=self._delay_s,
                until=lambda x: x != JobResult.GOOD
            )(check_all_services_up),
            counter_keeper=CountdownCounterKeeper(self._attempts),
            state_keeper=StateKeeper(),
        )


wait_all_services_up = WaitAllServicesUp



import os
from pathlib import Path


class Config:
    def __init__(self):
        self.project: str = os.environ.get('COMPOSE_PROJECT_NAME')
        self.non_stop_containers: list[str] = os.environ.get('NON_STOP_CONTAINERS').split(',')
        self.tmp_envs_path: Path = Path(os.environ.get('TMP_ENVS_DIRECTORY'))
        self.compose_files_path: Path = Path(os.environ.get('COMPOSE_FILES_DIRECTORY'))
        self.in_docker_project_root_path: Path = Path(os.environ.get('PROJECT_ROOT_DIRECTORY'))
        self.env_file_name: str = os.environ.get('TMP_ENVS_REGISTER_FILE')
        self.env_file_path: Path = self.tmp_envs_path / self.env_file_name
        self.host_project_root_directory: Path = Path(os.environ.get('HOST_PROJECT_ROOT_DIRECTORY'))
        self.env_tmp_directory: Path = Path(os.environ.get('TMP_ENVS_DIRECTORY'))
        self.host_env_tmp_directory: Path = Path(os.environ.get(
            'HOST_TMP_ENV_DIRECTORY',
            self.host_project_root_directory / self.env_tmp_directory.name
        ))
        self.docker_host = os.environ.get('DOCKER_HOST')
        self.compose_project_name = os.environ.get('COMPOSE_PROJECT_NAME')
        self.default_compose_files: str = os.environ.get('DEFAULT_COMPOSE_FILES')
        assert self.default_compose_files != '', 'unset DEFAULT_COMPOSE_FILES'
        self.debug_docker_compose_commands = bool(os.environ.get('DEBUG_DOCKER_COMPOSE_OUTPUT_TO_STDOUT', False))
        self.verbose_docker_compose_commands = bool(os.environ.get('VERBOSE_DOCKER_COMPOSE_OUTPUT_TO_STDOUT', True))
        self.verbose_docker_compose_ps_commands = bool(os.environ.get('VERBOSE_DOCKER_COMPOSE_PS_OUTPUT_TO_STDOUT', False))
        # TODO call it pre migration checks
        self.service_up_check_attempts = int(os.environ.get('PRE_MIGRATIONS_CHECK_SERVICE_UP_ATTEMPTS', 100))
        self.service_up_check_delay = int(os.environ.get('PRE_MIGRATIONS_CHECK_SERVICE_UP_CHECK_DELAY', 3))
        self.docker_compose_extra_exec_params = os.environ.get('DOCKER_COMPOSE_EXTRA_EXEC_PARAMS', '-T')



import pathlib

from setuptools import find_packages
from setuptools import setup


def find_required():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_version(filename='maxwelld/version'):
    return open(filename, "r").read().strip()


HERE = pathlib.Path(__file__).parent
README = open("README.md").read()
setup(
    name="maxwelld",
    version=get_version(),
    description="docker compose testing env orchestrator",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Yuriy Sagitov",
    author_email="pro100.ko10ok@gmail.com",
    python_requires=">=3.10.0",
    url="https://github.com/ko10ok/maxwelld",
    license="Apache-2.0",
    packages=find_packages(exclude=("tests",)),
    install_requires=find_required(),
    entry_points={},
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    package_data={
        'maxwelld': ['version'],
    },
)



import os
from _warnings import warn

from maxwelld.env_description.env_types import Environment


def setup_env_for_tests(env: Environment):
    updated_env = {}
    for service in env:
        for k, v in env[service].env.items():
            if k in updated_env and updated_env[k] != v:
                warn(
                    f'⚠️ env {k} tried to setup up multiple'
                    f' times with different values: {updated_env[k]} vs {v}'
                )
            updated_env[k] = v
            os.environ[k] = v



import os

from aiohttp import web

from maxwelld.server.commands import DC_EXEC_PATH
from maxwelld.server.commands import DC_LOGS_PATH
from maxwelld.server.commands import DC_UP_PATH
from maxwelld.server.commands import ENV_PATH
from maxwelld.server.commands import HEALTHCHECK_PATH
from maxwelld.server.commands import STATUS_PATH
from maxwelld.server.commands import UP_PATH
from maxwelld.server.handlers.dc_exec import dc_exec
from maxwelld.server.handlers.dc_logs import dc_logs
from maxwelld.server.handlers.dc_up import dc_up
from maxwelld.server.handlers.env import http_get_env
from maxwelld.server.handlers.healthcheck import healthcheck
from maxwelld.server.handlers.status import http_get_status
from maxwelld.server.handlers.up import up_compose

routes = web.RouteTableDef()

app = web.Application()
app.add_routes([
    web.post(DC_UP_PATH, dc_up),
    web.post(DC_EXEC_PATH, dc_exec),
    web.post(DC_LOGS_PATH, dc_logs),

    # ============================
    web.get(HEALTHCHECK_PATH, healthcheck),
    web.post(UP_PATH, up_compose),
    web.get(STATUS_PATH, http_get_status),
    web.get(ENV_PATH, http_get_env),
])


def run_server():
    web.run_app(app, port=os.environ.get('PORT', 80))



from maxwelld.client.maxwell_client import MaxwellDemonClient
from maxwelld.core.sequence_run_types import ComposeConfig
from maxwelld.core.service import MaxwellDemonService
from maxwelld.env_description.env_tools import off
from maxwelld.env_description.env_types import AsIs
from maxwelld.env_description.env_types import DEFAULT_ENV
from maxwelld.env_description.env_types import Env
from maxwelld.env_description.env_types import Environment
from maxwelld.env_description.env_types import Environments
from maxwelld.env_description.env_types import EventStage
from maxwelld.env_description.env_types import Handler
from maxwelld.env_description.env_types import Service
from maxwelld.vedro_plugin.plugin import DEFAULT_COMPOSE
from maxwelld.vedro_plugin.plugin import VedroMaxwell
from maxwelld.vedro_plugin.state_waiting import wait_all_services_up
from maxwelld.version import get_version

__version__ = get_version()
__all__ = (
    'MaxwellDemonClient', 'VedroMaxwell', 'DEFAULT_COMPOSE', 'ComposeConfig',
    'Environments', 'DEFAULT_ENV', 'Environment', 'Service', 'AsIs',
    'Handler', 'EventStage', 'off',
    'wait_all_services_up',
)



import base64
import pickle
from typing import Any


def base64_pickled(obj: Any) -> str:
    return base64.b64encode(pickle.dumps(obj)).decode('utf-8')


def debase64_pickled(data: str) -> Any:
    return pickle.loads(base64.b64decode(data))



import asyncio
from typing import TypedDict

from aiohttp import web
from aiohttp.web_request import Request

from maxwelld.client.types import EnvironmentId
from maxwelld.core.service import MaxwellDemonServiceManager
from maxwelld.helpers.bytes_pickle import debase64_pickled

UP_LOCK = asyncio.Lock()


class UpRequestParams(TypedDict):
    name: str
    config_template: str
    compose_files: str
    isolation: bool
    parallelism_limit: int
    force_restart: bool


class UpResponseParams(TypedDict):
    env_id: EnvironmentId


async def up_compose(request: Request) -> web.Response:
    params: UpRequestParams = await request.json()
    async with UP_LOCK:
        env_id, is_new = await MaxwellDemonServiceManager().get().up_compose(
            name=params['name'],
            config_template=debase64_pickled(params['config_template']),
            compose_files=params['compose_files'],
            isolation=params['isolation'],
            parallelism_limit=params['parallelism_limit'],
            force_restart=params['force_restart'],
        )
    return web.json_response(UpResponseParams(env_id=env_id), status=200)



import aiohttp
from aiohttp import ClientConnectorError
from rtry import retry

from maxwelld.client.types import EnvironmentId
from maxwelld.core.compose_data_types import ServicesComposeState
from maxwelld.errors.up import ServicesUpError
from maxwelld.env_description.env_types import Environment
from maxwelld.helpers.bytes_pickle import base64_pickled
from maxwelld.helpers.bytes_pickle import debase64_pickled
from maxwelld.server.commands import DC_EXEC_PATH
from maxwelld.server.commands import DC_LOGS_PATH
from maxwelld.server.commands import DC_UP_PATH
from maxwelld.server.commands import ENV_PATH
from maxwelld.server.commands import HEALTHCHECK_PATH
from maxwelld.server.commands import STATUS_PATH
from maxwelld.server.handlers.dc_exec import DcExecRequestParams
from maxwelld.server.handlers.dc_exec import DcExecResponseParams
from maxwelld.server.handlers.dc_logs import DcLogsRequestParams
from maxwelld.server.handlers.env import EnvRequestParams
from maxwelld.server.handlers.env import EnvResponseParams
from maxwelld.server.handlers.status import StatusResponseParams
from maxwelld.server.handlers.up import UpRequestParams
from maxwelld.server.handlers.up import UpResponseParams


class MaxwellDemonClient:
    def __init__(self, host, port=80):
        self._server_host = host
        self._server_port = port
        self._server_url = f'{self._server_host}:{self._server_port}'

    async def healthcheck(self):
        async with aiohttp.ClientSession() as session:
            url = f'{self._server_url}{HEALTHCHECK_PATH}'
            async with session.get(url) as response:
                assert response.status == 200, response
                state = await response.json()
                assert 'status' in state, response
                assert state['status'] == 'ok', response

    @retry(attempts=10, delay=1, swallow=ClientConnectorError)
    async def up(self, name, config_template: Environment, compose_files: str, isolation=None,
                 parallelism_limit=None, force_restart: bool = False) -> EnvironmentId:
        url = f'{self._server_url}{DC_UP_PATH}'
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=UpRequestParams(
                name=name,
                config_template=base64_pickled(config_template),
                compose_files=compose_files,
                isolation=isolation,
                parallelism_limit=parallelism_limit,
                force_restart=force_restart,
            )) as response:
                if response.status == 422:
                    raise ServicesUpError((await response.json())['error'])
                if response.status == 500:
                    raise ServicesUpError((await response.json())['error'])
                assert response.status == 200, response
                response_body = UpResponseParams(**await response.json())
                return response_body['env_id']

    @retry(attempts=5, delay=1, swallow=Exception)
    async def env(self, env_id: EnvironmentId) -> Environment:
        url = f'{self._server_url}{ENV_PATH}'
        async with aiohttp.ClientSession() as session:
            async with session.get(url, json=EnvRequestParams(id=env_id)) as response:
                assert response.status == 200, response
                response_body = EnvResponseParams(**await response.json())
                return debase64_pickled(response_body['env'])

    @retry(attempts=5, delay=1, swallow=Exception)
    async def status(self, env_id: EnvironmentId) -> ServicesComposeState:
        url = f'{self._server_url}{STATUS_PATH}'
        async with aiohttp.ClientSession() as session:
            async with session.get(url, json=EnvRequestParams(id=env_id)) as response:
                assert response.status == 200, response
                response_body = StatusResponseParams(**await response.json())
                return debase64_pickled(response_body['status'])

    async def exec(self, env_id: EnvironmentId, container: str, command: str) -> ServicesComposeState:
        url = f'{self._server_url}{DC_EXEC_PATH}'
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=DcExecRequestParams(
                env_id=env_id,
                container=container,
                command=command
            )) as response:
                assert response.status == 200, response
                response_body = DcExecResponseParams(**await response.json())
                return debase64_pickled(response_body['output'])

    async def logs(self, env_id: EnvironmentId, services: list[str]) -> dict[str, bytes]:
        url = f'{self._server_url}{DC_LOGS_PATH}'
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=DcLogsRequestParams(
                env_id=env_id,
                services=services,
            )) as response:
                assert response.status == 200, response
                response_body = DcExecResponseParams(**await response.json())
                return debase64_pickled(response_body['logs'])

    def list_current_in_flight_envs(self, *args, **kwargs):
        raise NotImplementedError()

    def list_services(self, *args, **kwargs):
        raise NotImplementedError()
